@@ -454,6 +454,7 @@ pub fn generate_compile_commands(
         num_parallel,
         executable: executable.to_owned(),
         arguments,
+        compilation_flag: parsed_args.compilation_flag.clone(),
         env_vars,
         cwd: cwd.to_owned(),
         host_compiler: host_compiler.clone(),
@@ -461,7 +462,20 @@ pub fn generate_compile_commands(
         output_file_name: output.file_name().unwrap().to_owned(),
     };
 
-    Ok((command, None, Cacheable::Yes))
+    Ok((
+        command,
+        None,
+        // Never assume the outer `nvcc` call is cacheable. We must decompose the nvcc call into
+        // its constituent subcommands with `--dryrun` and only cache the final build product.
+        //
+        // Always decomposing `nvcc --dryrun` is the only way to ensure caching nvcc invocations
+        // is fully sound, because the `nvcc -E` preprocessor output is not sufficient to detect
+        // all source code changes.
+        //
+        // Specifically, `nvcc -E` always defines __CUDA_ARCH__, which means changes to host-only
+        // code guarded by an `#ifndef __CUDA_ARCH__` will _not_ be captured in `nvcc -E` output.
+        Cacheable::No,
+    ))
 }
 
 #[derive(Clone, Debug)]
@@ -471,6 +485,7 @@ pub struct NvccCompileCommand {
     pub num_parallel: usize,
     pub executable: PathBuf,
     pub arguments: Vec<OsString>,
+    pub compilation_flag: OsString,
     pub env_vars: Vec<(OsString, OsString)>,
     pub cwd: PathBuf,
     pub host_compiler: NvccHostCompiler,
@@ -506,6 +521,7 @@ impl CompileCommandImpl for NvccCompileCommand {
             num_parallel,
             executable,
             arguments,
+            compilation_flag,
             env_vars,
             cwd,
             host_compiler,
@@ -516,6 +532,7 @@ impl CompileCommandImpl for NvccCompileCommand {
             creator,
             executable,
             arguments,
+            compilation_flag,
             cwd,
             temp_dir.as_path(),
             keep_dir.clone(),
@@ -612,6 +629,7 @@ async fn group_nvcc_subcommands_by_compilation_stage<T>(
     creator: &T,
     executable: &Path,
     arguments: &[OsString],
+    compilation_flag: &OsStr,
     cwd: &Path,
     tmp: &Path,
     keep_dir: Option<PathBuf>,
@@ -725,6 +743,19 @@ where
             ),
             // cicc and ptxas are cacheable
             Some("cicc") => {
+                match compilation_flag.to_str() {
+                    // Fix for CTK < 12.8:
+                    // If `nvcc` is invoked with `-c` (or any of its variants), remove the
+                    // `--gen_module_id_file` flag. In this mode, we instruct `cudafe++`
+                    // to generate this file, so cicc shouldn't generate it again.
+                    Some("-c") | Some("--compile") | Some("-dc") | Some("--device-c")
+                    | Some("-dw") | Some("--device-w") => {
+                        if let Some(idx) = args.iter().position(|x| x == &gen_module_id_file_flag) {
+                            args.splice(idx..(idx + 1), []);
+                        }
+                    }
+                    _ => {}
+                }
                 let group = device_compile_groups.get_mut(&args[args.len() - 3]);
                 (env_vars.clone(), Cacheable::Yes, group)
             }
@@ -739,7 +770,7 @@ where
                 });
                 (env_vars.clone(), Cacheable::Yes, group)
             }
-            // cudafe++ is not cacheable
+            // cudafe++ _must be_ cached, because the `.module_id` file is unique to each invocation (new in CTK 12.8)
             Some("cudafe++") => {
                 // Fix for CTK < 12.0:
                 // Add `--gen_module_id_file` if the cudafe++ args include `--module_id_file_name`
@@ -751,7 +782,7 @@ where
                 }
                 (
                     env_vars.clone(),
-                    Cacheable::No,
+                    Cacheable::Yes,
                     Some(&mut cuda_front_end_group),
                 )
             }
@@ -811,19 +842,28 @@ where
                         )
                     }
                 } else {
-                    // Returns Cacheable::Yes to indicate we _do_ want to run this host
-                    // compiler call through sccache (because it may be distributed),
-                    // but we _do not_ want to cache its output. The output file will
-                    // be cached as the result of the outer `nvcc` command. Caching
-                    // here would store the same object twice under two different hashes,
-                    // unnecessarily bloating the cache size.
+                    // Cache the host compiler calls, since we've marked the outer `nvcc` call
+                    // as non-cacheable. This ensures `sccache nvcc ...` _always_ decomposes the
+                    // nvcc call into its constituent subcommands with `--dryrun`, but only caches
+                    // the final build product once.
+                    //
+                    // Always decomposing `nvcc --dryrun` is the only way to ensure caching nvcc invocations
+                    // is fully sound, because the `nvcc -E` preprocessor output is not sufficient to detect
+                    // all source code changes.
+                    //
+                    // Specifically, `nvcc -E` always defines __CUDA_ARCH__, which means changes to host-only
+                    // code guarded by an `#ifndef __CUDA_ARCH__` will _not_ be captured in `nvcc -E` output.
                     (
                         env_vars
                             .iter()
                             .chain(
                                 [
-                                    // Do not cache host compiler calls
-                                    ("SCCACHE_NO_CACHE".into(), "true".into()),
+                                    // HACK: This compilation will look like a C/C++ compilation,
+                                    // but we want to report it in the stats as a CUDA compilation.
+                                    // The SccacheService API doesn't have a great way to specify this
+                                    // case, so we set a special envvar here that it can read when the
+                                    // compilation is finished.
+                                    ("__SCCACHE_THIS_IS_A_CUDA_COMPILATION__".into(), "".into()),
                                 ]
                                 .iter(),
                             )
@@ -1355,6 +1395,8 @@ counted_array!(pub static ARGS: [ArgInfo<gcc::ArgData>; _] = [
     flag!("--expt-relaxed-constexpr", PreprocessorArgumentFlag),
     flag!("--extended-lambda", PreprocessorArgumentFlag),
     flag!("--fatbin", DoCompilation),
+    take_arg!("--fdevice-time-trace", OsString, CanBeSeparated, TooHard),
+    take_arg!("--fdevice-time-trace=", OsString, Concatenated, TooHard),
     take_arg!("--generate-code", OsString, CanBeSeparated('='), PassThrough),
     flag!("--generate-dependencies-with-compile", NeedDepTarget),
     flag!("--generate-nonsystem-dependencies-with-compile", NeedDepTarget),
@@ -1375,6 +1417,8 @@ counted_array!(pub static ARGS: [ArgInfo<gcc::ArgData>; _] = [
     flag!("--save-temps", UnhashedFlag),
     take_arg!("--system-include", PathBuf, CanBeSeparated('='), PreprocessorArgumentPath),
     take_arg!("--threads", OsString, CanBeSeparated('='), Unhashed),
+    take_arg!("--time", OsString, CanBeSeparated, TooHard),
+    take_arg!("--time=", OsString, Concatenated, TooHard),
     take_arg!("--x", OsString, CanBeSeparated('='), Language),
 
     take_arg!("-Werror", OsString, CanBeSeparated('='), PreprocessorArgument),
@@ -1394,6 +1438,8 @@ counted_array!(pub static ARGS: [ArgInfo<gcc::ArgData>; _] = [
     flag!("-expt-relaxed-constexpr", PreprocessorArgumentFlag),
     flag!("-extended-lambda", PreprocessorArgumentFlag),
     flag!("-fatbin", DoCompilation),
+    take_arg!("-fdevice-time-trace", OsString, CanBeSeparated, TooHard),
+    take_arg!("-fdevice-time-trace=", OsString, Concatenated, TooHard),
     take_arg!("-gencode", OsString, CanBeSeparated('='), PassThrough),
     take_arg!("-isystem", PathBuf, CanBeSeparated('='), PreprocessorArgumentPath),
     flag!("-keep", UnhashedFlag),
@@ -1406,6 +1452,8 @@ counted_array!(pub static ARGS: [ArgInfo<gcc::ArgData>; _] = [
     flag!("-save-temps", UnhashedFlag),
     take_arg!("-t", OsString, CanBeSeparated, Unhashed),
     take_arg!("-t=", OsString, Concatenated, Unhashed),
+    take_arg!("-time", OsString, CanBeSeparated, TooHard),
+    take_arg!("-time=", OsString, Concatenated, TooHard),
     take_arg!("-x", OsString, CanBeSeparated('='), Language),
 ]);
 
@@ -1922,6 +1970,7 @@ mod test {
             ])
         );
     }
+
     #[test]
     fn test_parse_cant_cache_flags() {
         assert_eq!(
@@ -1949,5 +1998,52 @@ mod test {
             CompilerArguments::CannotCache("-M", None),
             parse_arguments_nvc(stringvec!["-x", "cu", "-c", "foo.c", "-o", "foo.o", "-M"])
         );
+
+        // nvcc arg parsing is very permissive, so all these are valid and should yield CannotCache
+        for arg in [
+            "-fdevice-time-trace",
+            "--fdevice-time-trace",
+            "-time",
+            "--time",
+        ] {
+            // {-,--}fdevice-time-trace
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_gcc(stringvec![arg])
+            );
+            // {-,--}fdevice-time-trace -
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_msvc(stringvec![arg, "-"])
+            );
+            // {-,--}fdevice-time-trace-
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_msvc(stringvec![format!("{arg}-")])
+            );
+            // {-,--}fdevice-time-trace flamegraph.json
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_nvc(stringvec![arg, "flamegraph.json"])
+            );
+        }
+
+        for arg_with_separator in [
+            "-fdevice-time-trace=",
+            "--fdevice-time-trace=",
+            "-time=",
+            "--time=",
+        ] {
+            // {-,--}fdevice-time-trace=-
+            assert_eq!(
+                CompilerArguments::CannotCache(arg_with_separator, None),
+                parse_arguments_msvc(stringvec![format!("{arg_with_separator}-")])
+            );
+            // {-,--}fdevice-time-trace=flamegraph.json
+            assert_eq!(
+                CompilerArguments::CannotCache(arg_with_separator, None),
+                parse_arguments_nvc(stringvec![format!("{arg_with_separator}flamegraph.json")])
+            );
+        }
     }
 }
